﻿/* IrbisProxy.cs
 */

#region Using directives

using System;
using System.Collections.Generic;
using System.Configuration;
using System.Diagnostics;
using System.Globalization;
using System.IO;
using System.Linq;
using System.Net;
using System.Net.Sockets;
using System.Runtime.Serialization;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading;

#endregion

namespace Always64
{
    // ====================================================================

    /// <summary>
    /// Простой планировщик задач. Выполняет поставленные задачи последовательно,
    /// одну за другой.
    /// Например, чтобы записи в лог-файлах не налезали друг на друга.
    /// </summary>
    /// <typeparam name="T">Тип аргумента, передаваемого задаче.</typeparam>
    class TaskQueue<T>
    {
        class TaskInfo
        {
            public Action<T> Action;

            public T Data;
        }

        // ====================================================================

        private readonly object _syncRoot;

        private readonly Queue<TaskInfo> _queue;

        private readonly AutoResetEvent _event;

        private volatile bool _stop;

        private volatile bool _executing;

        // ====================================================================

        public TaskQueue()
        {
            _syncRoot = new object();
            _queue = new Queue<TaskInfo>();
            _event = new AutoResetEvent(false);
            _stop = false;

            Thread workingThread = new Thread
                (
                    WorkingThread
                )
            {
                Name = "WorkingThread",
                IsBackground = true
            };
            workingThread.Start();
        }

        // ====================================================================

        public bool IsExecuting
        {
            get { return _executing; }
        }

        // ====================================================================

        public bool IsBusy
        {
            get { return (_queue.Count != 0) || _executing; }
        }

        // ====================================================================

        public void Stop()
        {
            _stop = true;
        }

        // ====================================================================

        void WorkingThread()
        {
            while (!_stop)
            {
                if (_stop)
                {
                    break;
                }

                TaskInfo task = null;

                lock (_syncRoot)
                {
                    if (_queue.Count != 0)
                    {
                        task = _queue.Dequeue();
                    }
                }

                if ( task == null )
                {
                    _event.WaitOne ();
                }
                else
                {
                    try
                    {
                        try
                        {
                            _executing = true;
                            task.Action(task.Data);
                        }
                        finally
                        {
                            _executing = false;
                            _event.Set ();
                        }
                    }
                    catch ( Exception ex )
                    {
                        IrbisProxy.WriteException(ex);
                        Debug.WriteLine ( ex );
                    }
                }
            }
        }

        // ====================================================================

        public void Enqueue
            (
                Action<T> action,
                T data
            )
        {
            TaskInfo task = new TaskInfo
            {
                Action = action,
                Data = data
            };

            lock (_syncRoot)
            {
                _queue.Enqueue(task);
            }

            _event.Set ();
        }
    }

    // ====================================================================

    /// <summary>
    /// Простой доступ к настройкам из App.config
    /// </summary>
    static class CM
    {
        public static string GetString
            (
                string name,
                string defaultValue
            )
        {
            string result = ConfigurationManager.AppSettings[name];
            if (string.IsNullOrEmpty(result))
            {
                result = defaultValue;
            }

            return result;
        }

        // ====================================================================

        public static int GetInt32
            (
                string name,
                int defaultValue
            )
        {
            string textValue = ConfigurationManager.AppSettings[name];
            if (string.IsNullOrEmpty(textValue))
            {
                return defaultValue;
            }

            int result;
            if (!int.TryParse(textValue, out result))
            {
                return defaultValue;
            }

            return result;
        }

        // ====================================================================

        public static bool GetBoolean
            (
                string name,
                bool defaultValue
            )
        {
            string textValue = ConfigurationManager.AppSettings[name];
            if (string.IsNullOrEmpty(textValue))
            {
                return defaultValue;
            }

            bool result;
            if (!bool.TryParse(textValue, out result))
            {
                return defaultValue;
            }

            return result;
        }

        // ====================================================================

        public static IPAddress GetAddress
            (
                string name,
                IPAddress defaultValue
            )
        {
            string textValue = ConfigurationManager.AppSettings[name];
            if (string.IsNullOrEmpty(textValue))
            {
                return defaultValue;
            }

            //Dns.GetHostAddresses(textValue);

            IPAddress result;
            if (!IPAddress.TryParse(textValue, out result))
            {
                return defaultValue;
            }

            return result;
        }
    }

    // ====================================================================

    static class Bits
    {
        public static byte[] Not(byte[] bytes)
        {
            return bytes.Select(b => (byte)~b).ToArray();
        }

        public static byte[] And(byte[] A, byte[] B)
        {
            int length = A.Length;
            byte[] result = new byte[length];
            for (int i = 0; i < length; i++)
            {
                result[i] = (byte)(A[i] & B[i]);
            }
            return result;
        }

        public static byte[] Or(byte[] A, byte[] B)
        {
            int length = A.Length;
            byte[] result = new byte[length];
            for (int i = 0; i < length; i++)
            {
                result[i] = (byte)(A[i] | B[i]);
            }
            return result;
        }

        public static bool GE(byte[] A, byte[] B)
        {
            int length = A.Length;
            int[] result = new int[length];
            for (int i = 0; i < length; i++)
            {
                int a = A[i], b = B[i];
                result[i] = a == b ? 0 : a < b ? 1 : -1;
            }

            return result
                .SkipWhile(c => c == 0)
                .FirstOrDefault() >= 0;
        }

        public static bool LE(byte[] A, byte[] B)
        {
            int length = A.Length;
            int[] result = new int[length];
            for (int i = 0; i < length; i++)
            {
                int a = A[i], b = B[i];
                result[i] = a == b ? 0 : a < b ? 1 : -1;
            }

            return result
                .SkipWhile(c => c == 0)
                .FirstOrDefault() <= 0;
        }

        public static byte[] GetBitMask(int sizeOfBuff, int bitLen)
        {
            var maskBytes = new byte[sizeOfBuff];
            var bytesLen = bitLen / 8;
            var bitsLen = bitLen % 8;
            for (int i = 0; i < bytesLen; i++)
            {
                maskBytes[i] = 0xff;
            }
            if (bitsLen > 0) maskBytes[bytesLen] = (byte)~Enumerable
                .Range(1, 8 - bitsLen).Select(n => 1 << n - 1)
                .Aggregate((a, b) => a | b);
            return maskBytes;
        }

    }

    // ====================================================================

    [Serializable]
    public sealed class IpRange
        : ISerializable
    {
        public IPAddress Begin { get; set; }

        public IPAddress End { get; set; }

        public IpRange()
        {
            Begin = new IPAddress(0);
            End = new IPAddress(0);
        }

        public IpRange
            (
                IPAddress begin,
                IPAddress end
            )
        {
            Begin = begin;
            End = end;
        }

        public IpRange
            (
                string ipRangeString
            )
        {
            if (string.IsNullOrEmpty(ipRangeString))
            {
                Begin = new IPAddress(0);
                End = new IPAddress(0);
                return;
            }
            if (ipRangeString == "*")
            {
                Begin = new IPAddress(0);
                End = new IPAddress(0xFFFFFFFF);
                return;
            }

            ipRangeString = ipRangeString.Replace
                (
                    " ",
                    string.Empty
                );

            Match match1 = Regex.Match
                (
                    ipRangeString,
                    @"^(?<addr>[\da-f\.:]+)/(?<maskLen>\d+)$",
                    RegexOptions.IgnoreCase
                );
            if (match1.Success)
            {
                byte[] addressBytes = IPAddress.Parse(match1.Groups["addr"].Value).GetAddressBytes();
                byte[] bitMask = Bits.GetBitMask(addressBytes.Length, int.Parse(match1.Groups["maskLen"].Value));
                addressBytes = Bits.And(addressBytes, bitMask);
                Begin = new IPAddress(addressBytes);
                End = new IPAddress(Bits.Or(addressBytes, Bits.Not(bitMask)));
                return;
            }

            Match match2 = Regex.Match
                (
                    ipRangeString,
                    @"^(?<addr>[\da-f\.:]+)$",
                    RegexOptions.IgnoreCase
                );
            if (match2.Success)
            {
                Begin = End = IPAddress.Parse(ipRangeString);
                return;
            }

            Match match3 = Regex.Match
                (
                    ipRangeString,
                    @"^(?<begin>[\da-f\.:]+)-(?<end>[\da-f\.:]+)$",
                    RegexOptions.IgnoreCase
                );
            if (match3.Success)
            {
                Begin = IPAddress.Parse(match3.Groups["begin"].Value);
                End = IPAddress.Parse(match3.Groups["end"].Value);
                return;
            }

            Match match4 = Regex.Match
                (
                    ipRangeString,
                    @"^(?<addr>[\da-f\.:]+)/(?<mask>[\da-f].:]+)$",
                    RegexOptions.IgnoreCase
                );
            if (match4.Success)
            {
                byte[] addressBytes = IPAddress.Parse(match4.Groups["addr"].Value).GetAddressBytes();
                byte[] bitMask = IPAddress.Parse(match4.Groups["mask"].Value).GetAddressBytes();
                addressBytes = Bits.And(addressBytes, bitMask);
                Begin = new IPAddress(addressBytes);
                End = new IPAddress(Bits.Or(addressBytes, Bits.Not(bitMask)));
                return;
            }

            throw new FormatException("Bad IP range string");
        }

        private IpRange
            (
                SerializationInfo info,
                StreamingContext context
            )
        {
            List<string> names = new List<string>();
            foreach (SerializationEntry entry in info)
            {
                names.Add(entry.Name);
            }

            Func<string, IPAddress> deserialize = name => names.Contains(name)
                ? IPAddress.Parse(info.GetValue(name, typeof(object)).ToString())
                : new IPAddress(0L);

            Begin = deserialize("Begin");
            End = deserialize("End");
        }

        public bool Contains
            (
                IPAddress address
            )
        {
            byte[] bytes = address.GetAddressBytes();
            return Bits.GE(Begin.GetAddressBytes(), bytes)
                   && Bits.LE(End.GetAddressBytes(), bytes);
        }

        public void GetObjectData
            (
                SerializationInfo info,
                StreamingContext context
            )
        {
            info.AddValue("Begin", Begin != null ? Begin.ToString() : string.Empty);
            info.AddValue("End", End != null ? End.ToString() : string.Empty);
        }

        public override string ToString()
        {
            string begin = Begin.ToString();
            string end = End.ToString();

            return (begin == end)
                ? begin
                : string.Concat
                    (
                        begin,
                        "-",
                        end
                    );
        }
    }

    // ====================================================================

    [Serializable]
    public sealed class IpFilter
    {
        public List<IpRange> Ranges { get { return _ranges; } }

        private readonly List<IpRange> _ranges = new List<IpRange>();

        public bool AllowAnyAddress { get; set; }

        public IpFilter()
        {
        }

        public IpFilter
            (
                string specification
            )
        {
            if (string.IsNullOrEmpty(specification)
                || (specification == "*"))
            {
                AllowAnyAddress = true;
                return;
            }

            string[] parts = specification.Split(';', ',');
            foreach (string part in parts)
            {
                if (!string.IsNullOrEmpty(part))
                {
                    Ranges.Add(new IpRange(part));
                }
            }
        }

        public bool Allowed
            (
                IPAddress address
            )
        {
            if (AllowAnyAddress)
            {
                return true;
            }

            foreach (IpRange range in Ranges)
            {
                if (range.Contains(address))
                {
                    return true;
                }
            }

            return false;
        }

        public bool Allowed
            (
                EndPoint endPoint
            )
        {
            IPEndPoint ip = endPoint as IPEndPoint;
            if (ReferenceEquals(ip, null))
            {
                return false;
            }
            return Allowed(ip.Address);
        }

        public override string ToString()
        {
            if (AllowAnyAddress)
            {
                return "*";
            }
            StringBuilder result = new StringBuilder();
            bool first = true;
            foreach (IpRange range in Ranges)
            {
                if (!first)
                {
                    result.Append(";");
                }
                result.Append(range);
                first = false;
            }
            return result.ToString();
        }
    }

    // ====================================================================

    /// <summary>
    /// Вся информация о выполненной команде.
    /// </summary>
    class TransactionInfo
    {
        /// <summary>
        /// Порядковый номер.
        /// </summary>
        public long Index { get; set; }

        /// <summary>
        /// Стадия, до которой дошло выполнение команды.
        /// </summary>
        public int Stage { get; set; }

        /// <summary>
        /// Момент поступления команды.
        /// </summary>
        public DateTime Moment { get; set; }

        /// <summary>
        /// Адрес, с которого пришла команда
        /// </summary>
        public EndPoint EndPoint { get; set; }

        /// <summary>
        /// Запрос клиента в необработанном виде
        /// (включая строки заголовка)
        /// </summary>
        public byte[] Request { get; set; }

        /// <summary>
        /// Ответ сервера в необработанном виде
        /// (включая строки заголовка)
        /// </summary>
        public byte[] Response { get; set; }

        /// <summary>
        /// Общая продолжительность обработки запроса
        /// со всеми пересылками туда-сюда
        /// (в миллисекундах).
        /// </summary>
        public long RoundtripDuration { get; set; }

        /// <summary>
        /// Продолжительность обработки запроса сервером
        /// ИРБИС (в миллисекундах).
        /// </summary>
        public long IrbisDuration { get; set; }
    }

    // ====================================================================

    /// <summary>
    /// Простой парсер заголовка запроса/ответа,
    /// разбирающий его на строки.
    /// </summary>
    class HeaderParser
    {
        private readonly byte[] _data;

        private int _offset;

        // ====================================================================

        public HeaderParser(byte[] data)
        {
            _data = data;
            _offset = 0;
        }

        // ====================================================================

        public string NextString()
        {
            int i;
            for (i = _offset; i < _data.Length; i++)
            {
                if (_data[i] == 0x0A)
                {
                    break;
                }
            }

            string result = Encoding.Default.GetString
                (
                    _data,
                    _offset,
                    i - _offset
                );

            _offset = i + 1;
            if (_offset > _data.Length)
            {
                _offset = _data.Length;
            }

            return result;
        }

        // ====================================================================

        public byte[] NextBytes()
        {
            int remainder = _data.Length - _offset;
            byte[] result = new byte[remainder];
            Array.Copy(_data, _offset, result, 0, remainder);
            return result;
        }

        // ====================================================================

        public string[] SplitLines
            (
                byte[] buffer,
                bool utf
            )
        {
            string[] result = null;
            try
            {
                Encoding encoding = utf
                                        ? Encoding.UTF8
                                        : Encoding.Default;
                string text = encoding.GetString(buffer);
                text = text.Replace ("\r\n", "\r");
                result = text.Split (new []{'\r', '\n'});
            }
            catch (Exception ex)
            {
                IrbisProxy.WriteException(ex);
            }

            return result;
        }
    }

    // ====================================================================

    /// <summary>
    /// Информация об ответе сервера.
    /// </summary>
    class ResponseInfo
    {
        /// <summary>
        /// Код команды.
        /// </summary>
        public string Command { get; set; }

        /// <summary>
        /// Идентификатор пользователя.
        /// </summary>
        public string UserID { get; set; }

        /// <summary>
        /// Порядковый номер команды.
        /// </summary>
        public string Index { get; set; }

        /// <summary>
        /// Зарезервировано.
        /// </summary>
        public string Reserv1 { get; set; }

        /// <summary>
        /// Зарезервировано.
        /// </summary>
        public string Reserv2 { get; set; }

        /// <summary>
        /// Зарезервировано.
        /// </summary>
        public string Reserv3 { get; set; }

        /// <summary>
        /// Зарезервировано.
        /// </summary>
        public string Reserv4 { get; set; }

        /// <summary>
        /// Зарезервировано.
        /// </summary>
        public string Reserv5 { get; set; }

        /// <summary>
        /// Зарезервировано.
        /// </summary>
        public string Reserv6 { get; set; }

        /// <summary>
        /// Зарезервировано.
        /// </summary>
        public string Reserv7 { get; set; }

        /// <summary>
        /// Прочие данные, например, найденные записи.
        /// </summary>
        public byte[] Data { get; set; }

        /// <summary>
        /// Прочие данные в текстовом виде.
        /// </summary>
        public string[] Lines { get; set; }

        // ====================================================================

        /// <summary>
        /// Код возврата (не для всех ответов имеет смысл).
        /// </summary>
        public string ReturnCode
        {
            get
            {
                if ((Lines == null) || (Lines.Length == 0))
                {
                    return null;
                }
                string result = Lines[0];
                result = OnlyNumbersAllowed ( result );
                return result;
            }
        }

        // ====================================================================

        private static string OnlyNumbersAllowed
            (
                string text
            )
        {
            if ( string.IsNullOrEmpty ( text ) )
            {
                return text;
            }

            int index = 0;
            if ( text [ 0 ] == '-' )
            {
                index++;
            }

            for ( ; index < text.Length; index++ )
            {
                if ( !char.IsDigit ( text, index ) )
                {
                    return null;
                }
            }

            return text;
        }

        // ====================================================================

        /// <summary>
        /// Разбор ответа сервера по полям заголовка.
        /// </summary>
        public static ResponseInfo Parse
            (
                byte[] buffer,
                bool useUtf
            )
        {
            HeaderParser parser = new HeaderParser(buffer);
            ResponseInfo result = new ResponseInfo
            {
                Command = parser.NextString(),
                UserID = parser.NextString(),
                Index = parser.NextString(),
                Reserv1 = parser.NextString(),
                Reserv2 = parser.NextString(),
                Reserv3 = parser.NextString(),
                Reserv4 = parser.NextString(),
                Reserv5 = parser.NextString(),
                Reserv6 = parser.NextString(),
                Reserv7 = parser.NextString(),
                Data = parser.NextBytes()
            };

            result.Lines = parser.SplitLines
                (
                    result.Data,
                    useUtf
                );

            return result;
        }

        // ====================================================================

        /// <summary>
        /// Текстовое представление заголовка ответа.
        /// </summary>
        /// <returns></returns>
        public override string ToString()
        {
            StringBuilder result = new StringBuilder();

            result.AppendFormat("Command: {0}", Command);
            result.AppendLine();
            result.AppendFormat("UserID: {0}", UserID);
            result.AppendLine();
            result.AppendFormat("Index: {0}", Index);
            //result.AppendLine();
            //result.AppendFormat("Reserv1: {0}", Reserv1);
            //result.AppendLine();
            //result.AppendFormat("Reserv2: {0}", Reserv2);
            //result.AppendLine();
            //result.AppendFormat("Reserv3: {0}", Reserv3);
            //result.AppendLine();
            //result.AppendFormat("Reserv4: {0}", Reserv4);
            //result.AppendLine();
            //result.AppendFormat("Reserv5: {0}", Reserv5);
            //result.AppendLine();
            //result.AppendFormat("Reserv6: {0}", Reserv6);
            //result.AppendLine();
            //result.AppendFormat("Reserv7: {0}", Reserv7);

            return result.ToString();
        }
    }

    // ====================================================================

    /// <summary>
    /// Запрос клиента в разобранном виде.
    /// </summary>
    class RequestInfo
    {
        /// <summary>
        /// Код команды.
        /// </summary>
        public string Command1 { get; set; }

        /// <summary>
        /// Тип АРМ.
        /// </summary>
        public string Workstation { get; set; }

        /// <summary>
        /// Код команды (повтор).
        /// </summary>
        public string Command2 { get; set; }

        /// <summary>
        /// Идентификатор клиента.
        /// </summary>
        public string UserID { get; set; }

        /// <summary>
        /// Порядковый номер запроса.
        /// </summary>
        public string Index { get; set; }

        /// <summary>
        /// Пароль.
        /// </summary>
        public string Password { get; set; }

        /// <summary>
        /// Логин.
        /// </summary>
        public string Username { get; set; }

        /// <summary>
        /// Зарезервировано.
        /// </summary>
        public string Reserv1 { get; set; }

        /// <summary>
        /// Зарезервировано.
        /// </summary>
        public string Reserv2 { get; set; }

        /// <summary>
        /// Зарезервировано.
        /// </summary>
        public string Reserv3 { get; set; }

        /// <summary>
        /// Прочие данные, например, поисковый запрос.
        /// </summary>
        public byte[] Data { get; set; }

        /// <summary>
        /// Прочие данные в текстовом виде.
        /// </summary>
        public string[] Lines { get; set; }

        // ====================================================================

        /// <summary>
        /// Разбор клиентского запроса по полям заголовка.
        /// </summary>
        public static RequestInfo Parse
            (
                byte[] buffer
            )
        {
            HeaderParser parser = new HeaderParser(buffer);
            
            // Пропускаем заголовок, содержащий общую длину пакета
            parser.NextString();
            
            RequestInfo result = new RequestInfo
            {
                Command1 = parser.NextString(),
                Workstation = parser.NextString(),
                Command2 = parser.NextString(),
                UserID = parser.NextString(),
                Index = parser.NextString(),
                Password = parser.NextString(),
                Username = parser.NextString(),
                Reserv1 = parser.NextString(),
                Reserv2 = parser.NextString(),
                Reserv3 = parser.NextString(),
                Data = parser.NextBytes()
            };

            result.Lines = parser.SplitLines
                (
                    result.Data,
                    !new[] { "8", "A", "B" }
                        .Contains ( result.Command1 )
                );

            return result;
        }

        /// <summary>
        /// Разбор клиентского запроса по полям заголовка.
        /// </summary>
        public static RequestInfo Parse
            (
                byte[] buffer,
                bool useUtf
            )
        {
            HeaderParser parser = new HeaderParser(buffer);

            // Пропускаем заголовок, содержащий общую длину пакета
            parser.NextString();

            RequestInfo result = new RequestInfo
            {
                Command1 = parser.NextString(),
                Workstation = parser.NextString(),
                Command2 = parser.NextString(),
                UserID = parser.NextString(),
                Index = parser.NextString(),
                Password = parser.NextString(),
                Username = parser.NextString(),
                Reserv1 = parser.NextString(),
                Reserv2 = parser.NextString(),
                Reserv3 = parser.NextString(),
                Data = parser.NextBytes()
            };

            result.Lines = parser.SplitLines
                (
                    result.Data,
                    useUtf
                );

            return result;
        }

        // ====================================================================

        /// <summary>
        /// Текстовое представление запроса.
        /// </summary>
        /// <returns></returns>
        public override string ToString()
        {
            StringBuilder result = new StringBuilder();

            result.AppendFormat("Command1: {0}", Command1);
            result.AppendLine();
            result.AppendFormat("Workstation: {0}", Workstation);
            result.AppendLine();
            result.AppendFormat("Command2: {0}", Command2);
            result.AppendLine();
            result.AppendFormat("UserID: {0}", UserID);
            result.AppendLine();
            result.AppendFormat("Index: {0}", Index);
            result.AppendLine();
            result.AppendFormat("Password: {0}", Password);
            result.AppendLine();
            result.AppendFormat("Username: {0}", Username);
            //result.AppendLine();
            //result.AppendFormat("Reserv1: {0}", Reserv1);
            //result.AppendLine();
            //result.AppendFormat("Reserv2: {0}", Reserv2);
            //result.AppendLine();
            //result.AppendFormat("Reserv3: {0}", Reserv3);

            return result.ToString();
        }
    }

    // ====================================================================

    /// <summary>
    /// Рабочий код прокси.
    /// </summary>
    static class IrbisProxy
    {
        public static Version Version = new Version( 1, 0, 10, 26 );

        /// <summary>
        /// Признак остановки.
        /// </summary>
        public static volatile bool Stop;

        /// <summary>
        /// Нагрузка на сервер: количество одновременных
        /// соединений в данный момент.
        /// </summary>
        public static int Load;

        /// <summary>
        /// Простая последовательная очередь задач
        /// на запись в логи.
        /// </summary>
        private static TaskQueue<TransactionInfo> Engine;

        /// <summary>
        /// Прослушивающий сокет.
        /// </summary>
        private static Socket ListenerSocket;

        /// <summary>
        /// Последовательный номер обрабатываемой транзакции
        /// «запрос-ответ».
        /// </summary>
        private static long TransactionIndex;

        /// <summary>
        /// Таймаут приема данных в миллисекундах.
        /// 0 — не устанавливать.
        /// -1 – установить равным бесконечности.
        /// </summary>
        private static int ReceiveTimeout;

        /// <summary>
        /// Таймаут посылки данных в миллисекундах.
        /// 0 – не устанавливать.
        /// -1 – установить равным бесконечности.
        /// </summary>
        private static int SendTimeout;

        /// <summary>
        /// Номер прослушивающего порта.
        /// </summary>
        private static int LocalPort;

        /// <summary>
        /// Длина очереди прослушивающего сокета.
        /// </summary>
        private static int Backlog;

        /// <summary>
        /// Адрес сервера ИРБИС64.
        /// </summary>
        private static IPAddress RemoteIP;

        /// <summary>
        /// Порт, прослушиваемый сервером ИРБИС64.
        /// </summary>
        private static int RemotePort;

        /// <summary>
        /// Имя файла, в который выводится дамп.
        /// Если не задано, то используется консоль.
        /// </summary>
        private static string DumpTo;

        /// <summary>
        /// Выводить в консоль общую информацию о транзакции?
        /// </summary>
        private static bool DumpGeneralInfo;

        /// <summary>
        /// Выводить в консоль заголовок запроса?
        /// </summary>
        private static bool DumpRequestHeader;

        /// <summary>
        /// Выводить в консоль тело запроса?
        /// </summary>
        private static bool DumpRequestBody;

        /// <summary>
        /// Выводить в консоль заголовок ответа?
        /// </summary>
        private static bool DumpResponseHeader;

        /// <summary>
        /// Выводить в консоль тело ответа?
        /// </summary>
        private static bool DumpResponseBody;

        /// <summary>
        /// Выводить ли сведения об ошибках?
        /// </summary>
        private static bool DumpErrors;

        /// <summary>
        /// Писать лог подключений?
        /// </summary>
        private static bool WriteAccessLog;

        /// <summary>
        /// Имя файла лога подключений.
        /// </summary>
        private static string AccessLogFileName;

        /// <summary>
        /// Писать лог поисковых запросов?
        /// </summary>
        private static bool WriteSearchLog;

        /// <summary>
        /// Имя файла лога поисковых запросов.
        /// </summary>
        private static string SearchLogFileName;

        /// <summary>
        /// Имя файла с логами ошибок.
        /// </summary>
        private static string ErrorLogFileName;

        /// <summary>
        /// Записывать сетевую активность.
        /// </summary>
        private static bool WriteNetworkLog;

        /// <summary>
        /// Имя файла лога сетевой активности.
        /// </summary>
        private static string NetworkLogFileName;

        /// <summary>
        /// Команда, запускаемая для сообщения
        /// сисадмину о возникших проблемах.
        /// </summary>
        private static string WakeupCommand;

        /// <summary>
        /// Аргументы, передаваемые программе
        /// для извещения сисадмина.
        /// </summary>
        private static string WakeupArguments;

        /// <summary>
        /// Промежуток (в секундах) между
        /// последовательными запусками
        /// команды для сообщения сисадмину.
        /// </summary>
        private static int WakeupInterval;

        /// <summary>
        /// Добавлять "* HAVE=1" в запросы
        /// АРМ "Читатель"
        /// </summary>
        private static bool AddHave;

        /// <summary>
        /// Фильтр входящих соединений.
        /// </summary>
        private static IpFilter IpFilter;

        // ====================================================================

        public static void DumpSettings()
        {
            DrawLine();
            DrawLine();
            WriteDump("STARTED={0}", DateTime.Now);
            WriteDump("PROXY={0}", Version);
            WriteDump("CLR={0}", Environment.Version);
            WriteDump("ENVIROMENT={0}", Environment.OSVersion);
            WriteDump("USED={0}", Environment.WorkingSet);
            WriteDump("receive-timeout={0}", ReceiveTimeout);
            WriteDump("send-timeout={0}", SendTimeout);
            WriteDump("local-port={0}",LocalPort);
            WriteDump("backlog={0}",Backlog);
            WriteDump("remote-ip={0}",RemoteIP);
            WriteDump("remote-port={0}",RemotePort);
            WriteDump("dump-to={0}",DumpTo);
            WriteDump("dump-general-info={0}",DumpGeneralInfo);
            WriteDump("dump-request-header={0}",DumpRequestHeader);
            WriteDump("dump-request-body={0}",DumpRequestBody);
            WriteDump("dump-response-header={0}",DumpResponseHeader);
            WriteDump("dump-response-body={0}",DumpResponseBody);
            WriteDump("dump-errors={0}",DumpErrors);
            WriteDump("write-access-log={0}",WriteAccessLog);
            WriteDump("access-log-file-name={0}",AccessLogFileName);
            WriteDump("write-search-log={0}",WriteSearchLog);
            WriteDump("search-log-file-name={0}",SearchLogFileName);
            WriteDump("error-log-file-name={0}",ErrorLogFileName);
            WriteDump("write-network-log={0}",WriteNetworkLog);
            WriteDump("network-log-file-name={0}",NetworkLogFileName);
            WriteDump("wakeup-command={0}",WakeupCommand);
            WriteDump("wakeup-arguments={0}",WakeupArguments);
            WriteDump("wakeup-interval={0}",WakeupInterval);
            WriteDump("add-have={0}",AddHave);
            WriteDump("ip-filter={0}",IpFilter);
            DrawLine();
        }

        // ====================================================================

        /// <summary>
        /// Инициализация: считываем конфигурацию.
        /// </summary>
        public static void Initialize()
        {
            Engine = new TaskQueue<TransactionInfo>();

            ReceiveTimeout = CM.GetInt32
                (
                    "receive-timeout",
                    0
                );
            SendTimeout = CM.GetInt32
                (
                    "send-timeout",
                    0
                );
            LocalPort = CM.GetInt32
                (
                    "local-port",
                    5555
                );
            Backlog = CM.GetInt32
                (
                    "backlog",
                    10
                );
            RemoteIP = CM.GetAddress
                (
                    "remote-ip",
                    IPAddress.Loopback
                );
            RemotePort = CM.GetInt32
                (
                    "remote-port",
                    6666
                );
            DumpTo = CM.GetString
                (
                    "dump-to",
                    null
                );
            DumpGeneralInfo = CM.GetBoolean
                (
                    "dump-general-info",
                    false
                );
            DumpRequestHeader = CM.GetBoolean
                (
                    "dump-request-header",
                    false
                );
            DumpRequestBody = CM.GetBoolean
                (
                    "dump-request-body",
                    false
                );
            DumpResponseHeader = CM.GetBoolean
                (
                    "dump-response-header",
                    false
                );
            DumpResponseBody = CM.GetBoolean
                (
                    "dump-response-body",
                    false
                );
            DumpErrors = CM.GetBoolean
                (
                    "dump-errors",
                    false
                );
            WriteAccessLog = CM.GetBoolean
                (
                    "write-access-log",
                    false
                );
            AccessLogFileName = CM.GetString
                (
                    "access-log-file-name",
                    @"C:\access.log"
                );
            WriteSearchLog = CM.GetBoolean
                (
                    "write-search-log",
                    false
                );
            SearchLogFileName = CM.GetString
                (
                    "search-log-file-name",
                    @"C:\search.log"
                );
            ErrorLogFileName = CM.GetString
                (
                    "error-log-file-name",
                    @"C:\errors.log"
                );
            WriteNetworkLog = CM.GetBoolean
                (
                    "write-network-log",
                    false
                );
            NetworkLogFileName = CM.GetString
                (
                    "network-log-file-name",
                    @"C:\network.log"
                );
            WakeupCommand = CM.GetString
                (
                    "wakeup-command",
                    string.Empty
                );
            WakeupArguments = CM.GetString
                (
                    "wakeup-arguments",
                    null
                );
            WakeupInterval = CM.GetInt32
                (
                    "wakeup-interval",
                    300
                );
            AddHave = CM.GetBoolean
                (
                    "add-have",
                    false
                );
            IpFilter = new IpFilter
                (
                    CM.GetString
                    (
                        "ip-filter",
                        "*"                        
                    )
                );

            DumpSettings();
        }

        // ====================================================================

        public static void WriteException
            (
                Exception exception
            )
        {
            if (!DumpErrors
                || string.IsNullOrEmpty(ErrorLogFileName))
            {
                return;
            }

            try
            {
                string fileName = GetTodayFile(ErrorLogFileName);
                using (StreamWriter writer = new StreamWriter
                    (
                    fileName,
                    true
                    ))
                {
                    writer.WriteLine("------------------------------------------------------------");
                    writer.WriteLine("DATE, TIME: {0}", DateTime.Now);
                    writer.WriteLine(exception.ToString());
                    writer.WriteLine();
                }
            }
            catch
            {
                DumpErrors = false;
            }
        }

        // ====================================================================

        public static void LogNetwork
            (
                int stage,
                string format,
                params object[] args
            )
        {
            if (!WriteNetworkLog
                || string.IsNullOrEmpty(NetworkLogFileName))
            {
                return;
            }

            try
            {
                string fileName = GetTodayFile(NetworkLogFileName);
                using (StreamWriter writer = new StreamWriter
                    (
                        fileName,
                        true
                    ))
                {
                    writer.Write
                        (
                            "{0}: {1}: [{2:00}]: ", 
                            DateTime.Now,
                            Thread.CurrentThread.ManagedThreadId,
                            stage
                        );
                    writer.WriteLine
                        (
                            format, 
                            args
                        );
                }
            }
            catch
            {
                WriteNetworkLog = false;
            }
        }

        // ====================================================================

        public static void WriteDump
            (
                string format,
                params object[] args
            )
        {
            if ( string.IsNullOrEmpty ( DumpTo ) )
            {
                Console.WriteLine( format, args );
                return;
            }

            try
            {
                string fileName = GetTodayFile(DumpTo);
                using (StreamWriter writer = new StreamWriter
                    (
                        fileName,
                        true
                    ))
                {
                    writer.WriteLine(format, args);
                }
            }
            catch
            {
                DumpTo = null;
            }
        }

        // ====================================================================

        static void DrawLine()
        {
            WriteDump
                (
                    "------------------------------------------------------------"
                );
        }

        /// <summary>
        /// Настройка сокета (таймаут подключения и посылка
        /// без задержек.
        /// </summary>
        static void SetupSocket
            (
                Socket socket
            )
        {
            if (SendTimeout != 0)
            {
                socket.SendTimeout = SendTimeout;
            }
            if (ReceiveTimeout != 0)
            {
                socket.ReceiveTimeout = ReceiveTimeout;
            }

            socket.NoDelay = true;
        }

        // ====================================================================

        /// <summary>
        /// Прочитать из сокета не менее указанного количества байт.
        /// </summary>
        /// <returns>Удалось ли?</returns>
        static bool ReceiveExact
            (
                Socket socket,
                MemoryStream memory,
                int howMany
            )
        {
            try
            {
                byte[] buffer = new byte[1000];
                int toRead = Math.Min(buffer.Length, howMany);
                while (toRead > 0)
                {
                    int readed = socket.Receive
                        (
                            buffer,
                            0,
                            toRead,
                            SocketFlags.None
                        );
                    if (readed <= 0)
                    {
                        return false;
                    }
                    memory.Write
                        (
                            buffer,
                            0,
                            readed
                        );
                    howMany -= readed;
                    toRead = Math.Min(buffer.Length, howMany);
                }
            }
            catch (Exception ex)
            {
                WriteException(ex);
                Debug.WriteLine(ex);
                return false;
            }
            return true;
        }

        // ====================================================================

        /// <summary>
        /// Прочитать из сокета максимально возможное число байт.
        /// </summary>
        /// <returns>Удалось ли?</returns>
        static bool ReceiveAll
            (
                Socket socket,
                MemoryStream memory
            )
        {
            byte[] buffer = new byte[1000];
            try
            {
                int readed;
                do
                {
                    readed = socket.Receive(buffer);
                    if (readed > 0)
                    {
                        memory.Write(buffer, 0, readed);
                    }
                } while (readed > 0);
            }
            catch (Exception ex)
            {
                WriteException(ex);
                Debug.WriteLine(ex);
                return false;
            }

            return (memory.Length != 0);
        }

        // ====================================================================

        /// <summary>
        /// Отправить в сеть байты.
        /// </summary>
        /// <returns>Удалось ли?</returns>
        private static bool SendAll
            (
                Socket socket,
                byte[] buffer
            )
        {
            try
            {
                return (socket.Send(buffer) == buffer.Length);
            }
            catch (Exception ex)
            {
                WriteException(ex);
                Debug.WriteLine(ex);
                return false;
            }
        }

        // ====================================================================

        static string GetTodayFile
            (
                string path
            )
        {
            string directory = Path.GetDirectoryName ( path );
            string name = Path.GetFileNameWithoutExtension ( path );
            string ext = Path.GetExtension ( path );
            string result = string.Format
                (
                    "{0}{1}{2}{3:-yyyy-MM-dd}{4}",
                    directory,
                    Path.DirectorySeparatorChar,
                    name,
                    DateTime.Today,
                    ext
                );
            return result;
        }

        // ====================================================================

        /// <summary>
        /// Пишем в лог подключений.
        /// </summary>
        static void DoAccessLog
            (
                TransactionInfo transaction,
                RequestInfo request,
                ResponseInfo response
            )
        {
            try
            {
                using (StreamWriter writer = new StreamWriter
                    (
                        GetTodayFile ( AccessLogFileName ),
                        true
                    ))
                {
                    writer.WriteLine
                        (
                            "{0}\t{1}\t{2}\t{3}\t{4}\t{5}\t{6}\t{7}\t{8}\t{9}\t{10}"
                            + "\t{11}\t{12}\t{13}\t{14}",
                            transaction.Index,             // Порядковый номер транзакции
                            transaction.Moment,            // Момент начала транзакции
                            transaction.Stage,             // Стадия транзакции
                            transaction.EndPoint,          // Адрес клиента
                            request.UserID,                // Идентификатор клиента
                            request.Workstation,           // Код АРМ
                            request.Index,                 // Номер команды
                            request.Username,              // Логин пользователя
                            request.Command1,              // Команда
                            Load,                          // Нагрузка на сервер: кол-во соединений
                            transaction.IrbisDuration,     // Длительность обработки команды ИРБИС
                            transaction.RoundtripDuration, // Общая длительность обработки команды
                            transaction.Request.Length,    // Длина запроса в байтах
                            transaction.Response.Length,   // Длина ответа в байтах
                            response.ReturnCode            // Код возврата (не для всех команд)
                        );
                }
            }
            catch ( Exception ex )
            {
                WriteException(ex);
                Debug.WriteLine ( ex );
            }
        }

        // ====================================================================

        static T SafeGet < T > ( T[] array, int index )
        {
            try
            {
                return array [ index ];
            }
            catch
            {
                return default(T);
            }
        }

        // ====================================================================

        /// <summary>
        /// Пишем в лог поисковых запросов.
        /// </summary>
        static void DoSearchLog
            (
                TransactionInfo transaction,
                RequestInfo request,
                ResponseInfo response
            )
        {
            try
            {
                if (( request.Command1 != "K" )
                    || ( request.Workstation != "R")
                    || ( transaction.Stage != 13 ))
                {
                    return;
                }

                using (StreamWriter writer = new StreamWriter
                    (
                        GetTodayFile ( SearchLogFileName ),
                        true
                    ))
                {
                    writer.WriteLine
                        (
                            "{0}\t{1}\t{2}\t{3}\t{4}\t{5}\t{6}\t{7}\t{8}\t{9}\t{10}",
                            transaction.Index,             // Порядковый номер транзакции
                            transaction.Moment,            // Момент начала транзакции
                            transaction.EndPoint,          // Адрес клиента
                            request.Username,              // Логин пользователя
                            transaction.IrbisDuration,     // Длительность обработки команды
                            response.ReturnCode,           // Код возврата
                            SafeGet ( request.Lines, 0 ),  // База данных
                            SafeGet ( request.Lines, 1 ),  // Поисковое выражение
                            SafeGet ( request.Lines, 2 ),  // Число запрашиваемых записей
                            SafeGet ( request.Lines, 3 ),  // Номер первой возвращаемой записи
                            SafeGet ( response.Lines, 1 )  // Количество найденных
                        );
                }
            }
            catch ( Exception ex )
            {
                WriteException(ex);
                Debug.WriteLine ( ex );
            }
        }

        // ====================================================================

        static bool OneOf < T > ( this T one, params T[] many )
        {
            return many.Contains ( one );
        }

        // ====================================================================

        /// <summary>
        /// Отработка логирования (рабочая рутина).
        /// </summary>
        static void LogRoutine
            (
                TransactionInfo transaction
            )
        {
            if (DumpGeneralInfo)
            {
                DrawLine();
                WriteDump 
                    (
                        "{0}) {1} from {2} at stage {3}",
                        transaction.Index,
                        transaction.Moment,
                        transaction.EndPoint,
                        transaction.Stage
                    );
                WriteDump
                    (
                        "Request {0} byte(s), response {1} byte(s), "
                        + "duration {2} ms",
                        transaction.Request.Length,
                        transaction.Response.Length,
                        transaction.IrbisDuration
                    );
            }

            RequestInfo requestInfo = RequestInfo.Parse
                (
                    transaction.Request
                );

            if (DumpRequestHeader)
            {
                DrawLine();
                WriteDump 
                    ( 
                        "Request header:{0}{1}", 
                        Environment.NewLine,
                        requestInfo 
                    );
            }

            if (DumpRequestBody)
            {
                DrawLine();
                WriteDump ( "Request body:" );
                foreach ( string line in requestInfo.Lines )
                {
                    WriteDump ( "{0}", line );
                }
            }

            ResponseInfo responseInfo = ResponseInfo.Parse 
                ( 
                    transaction.Response,
                    !requestInfo.Command1.OneOf ( "A", "L" )
                );

            if (DumpResponseHeader)
            {
                DrawLine();
                WriteDump 
                    ( 
                        "Response header:{0}{1}",
                        Environment.NewLine,
                        responseInfo 
                    );
            }

            if (DumpResponseBody)
            {
                DrawLine();
                WriteDump ( "Response body:" );
                foreach ( string line in responseInfo.Lines )
                {
                    WriteDump ( "{0}", line );
                }
            }

            if (WriteAccessLog)
            {
                DoAccessLog
                    (
                        transaction,
                        requestInfo,
                        responseInfo
                    );
            }

            if (WriteSearchLog)
            {
                DoSearchLog
                    (
                        transaction,
                        requestInfo,
                        responseInfo
                    );
            }
        }

        // ====================================================================

        /// <summary>
        /// Отработка логирования (постановка в очередь).
        /// </summary>
        static void LogTransaction
            (
                TransactionInfo transaction
            )
        {
            Engine.Enqueue
                (
                    LogRoutine,
                    transaction
                );
        }

        // ====================================================================

        static void WakeupThread()
        {
            if (!string.IsNullOrEmpty(WakeupCommand))
            {
                try
                {
                    using (Process.Start
                        (
                            WakeupCommand,
                            WakeupArguments
                        ))
                    {
                        // Nothing to do here
                    }
                }
// ReSharper disable EmptyGeneralCatchClause
                catch
// ReSharper restore EmptyGeneralCatchClause
                {
                }
            }
        }

        // ====================================================================

        // Синхронизация для обращения к LastWakeup
        private static readonly object WakeupSync = new object();

        // Время предыдущего сообщения системному администратору
        private static DateTime LastWakeup;

        // ====================================================================

        // ReSharper disable UnusedParameter.Local
        static void WakeTheSystemAdministrator
            (
                TransactionInfo transaction
            )
        {
            lock (WakeupSync)
            {
                DateTime now = DateTime.Now;
                if ((now - LastWakeup).TotalSeconds < WakeupInterval)
                {
                    return;
                }
                LastWakeup = now;
            }

            Thread wakeupThread = new Thread(WakeupThread)
                {
                    IsBackground = true,
                    Name = "WakeupThread"
                };
            wakeupThread.Start();
        }
        // ReSharper restore UnusedParameter.Local

        // ====================================================================

        // ReSharper disable UnusedParameter.Local
        static bool CheckTheRequest
            (
                TransactionInfo transaction
            )
        {
            // TODO проверить, имеет ли право клиент посылать подобные запросы
            // TODO проверить, не битый ли пакет с запросом
            return true;
        }
        // ReSharper restore UnusedParameter.Local

        // ====================================================================

        // Добавляем "* HAVE=1" к запросам от АРМ Читатель
        // ReSharper disable UnusedParameter.Local
        static void TransformTheRequest
            (
                TransactionInfo transaction,
                ref byte[] bytes
            )
        {
            if (!AddHave)
            {
                return;
            }

            RequestInfo request = RequestInfo.Parse
                (
                    bytes,
                    false
                );
            if ((request.Workstation != "R")
                || (request.Command1 != "K")
                )
            {
                return;
            }

            try
            {

                string database = SafeGet(request.Lines, 0);
                if (string.Compare(database, "IBIS",
                    StringComparison.InvariantCultureIgnoreCase) != 0)
                {
                    return;
                }
                string query = SafeGet(request.Lines, 1);
                string rewritten = string.Format
                    (
                        "({0}) * HAVE=1",
                        query
                    );
                MemoryStream bodyStream = new MemoryStream(bytes.Length + 10);
                MemoryStream headStream = new MemoryStream();
                using (TextWriter ansi = new StreamWriter(bodyStream, Encoding.Default))
                {
                    ansi                        
                        .AppendLine(request.Command1)
                        .AppendLine(request.Workstation)
                        .AppendLine(request.Command2)
                        .AppendLine(request.UserID)
                        .AppendLine(request.Index)
                        .AppendLine(request.Password)
                        .AppendLine(request.Username)
                        .AppendLine(request.Reserv1)
                        .AppendLine(request.Reserv2)
                        .AppendLine(request.Reserv3)
                        .AppendLine(database)
                        .AppendLine(rewritten);
                    for (int i = 2; i < request.Lines.Length; i++)
                    {
                        ansi.AppendLine(SafeGet(request.Lines, i));
                    }
                }

                byte[] bodyBytes = bodyStream.ToArray();

                using (TextWriter ansi = new StreamWriter(headStream, Encoding.Default))
                {
                    ansi
                        .AppendLine(bodyBytes.Length.ToString(CultureInfo.InvariantCulture));
                }

                byte[] headBytes = headStream.ToArray();
                
                MemoryStream resultStream = new MemoryStream();
                resultStream.Write(headBytes, 0, headBytes.Length);
                resultStream.Write(bodyBytes, 0, bodyBytes.Length);

                bytes = resultStream.ToArray();
            }
            catch (Exception ex)
            {
                WriteException(ex);   
                AddHave = false;
            }
        }
        // ReSharper restore UnusedParameter.Local

        // ====================================================================

        // ReSharper disable UnusedParameter.Local
        static bool CheckTheResponse
            (
                TransactionInfo transaction
            )
        {
            // TODO проверить, не битый ли пакет
            // TODO подмена ответа
            return true;
        }
        // ReSharper restore UnusedParameter.Local

        // ====================================================================

        /// <summary>
        /// Отработка подключения.
        /// </summary>
        /// <param name="data"></param>
        static void HandlerRoutine
            (
                object data
            )
        {
            TransactionInfo transaction = new TransactionInfo
            {
                Moment = DateTime.Now,
                Index = Interlocked.Increment(ref TransactionIndex),
                Request = new byte[0],
                Response = new byte[0]
            };

            // Замеряем общее время, затраченное на обработку
            // со всеми пересылками туда-сюда
            Stopwatch roundtripStopwatch = new Stopwatch();
            roundtripStopwatch.Start ();

            // Сокет, связанный с подключившимся клиентом
            Socket downSocket = (Socket)data;
            
            LogNetwork
                (
                    1,
                    "Got DownSocket: {0}",
                    downSocket.RemoteEndPoint
                );

            Interlocked.Increment(ref Load);

            try
            {
                try
                {
                    transaction.Stage++; // 1
                    transaction.EndPoint = downSocket.RemoteEndPoint;

                    if (!IpFilter.Allowed(transaction.EndPoint))
                    {
                        LogNetwork(1, "Not allowed remote endpoint");
                        return;
                    }

                    LogNetwork(2, "Setup DownSocket");
                    SetupSocket(downSocket);

                    // Считываем первые 10 байт, чтобы определить,
                    // сколько надо будет прочитать еще.
                    MemoryStream requestStorage = new MemoryStream();
                    LogNetwork(3, "Begin ReceiveExact 1st time from DownSocket" );
                    if (!ReceiveExact
                        (
                            downSocket,
                            requestStorage,
                            10
                        ))
                    {
                        LogNetwork(4, "Fail ReceiveExact 1st time from DownSocket");

                        // Если не удалось прочитать первые 10 байт,
                        // не стоит и пытаться обрабатывать такого клиента.
                        // Не будем беспокоить сервер по таким мелочам.
                        return;
                    }
                    LogNetwork(5, "End ReceiveExact 1st time from DownSocket");

                    transaction.Stage++; // 2

                    // Ищем первый символ перевода строки 0x0A
                    byte[] buffer1 = requestStorage.ToArray();
                    int offset = Array.IndexOf<byte>(buffer1, 0x0A);
                    if (offset <= 0)
                    {
                        // Если не нашли, значит, пришел мусор,
                        // который нельзя передавать серверу.
                        return;
                    }

                    transaction.Stage++; // 3

                    // Вычисляем, сколько байт еще необходимо прочитать
                    string lengthText = Encoding.ASCII.GetString
                        (
                            buffer1,
                            0,
                            offset
                        );
                    int toRead;
                    try
                    {
                        int firstLength = ((int) requestStorage.Length) - offset - 1;
                        toRead = int.Parse(lengthText) - firstLength;
                    }
                    catch (Exception ex)
                    {
                        WriteException(ex);

                        // Первая строка – не число, не будем связываться
                        // с такими пакетами данных
                        return;
                    }
                    if ((toRead < 1) || (toRead > 1000*1000*1000))
                    {
                        // Однако, фигня пришла, не будем связываться
                        // с такими пакетами данных.
                        return;
                    }
                    LogNetwork(6, "Begin ReceiveExact 2nd time from DownSocket");
                    if (!ReceiveExact
                        (
                            downSocket,
                            requestStorage,
                            toRead
                        ))
                    {
                        LogNetwork(7, "Fail ReceiveExact 2nd time from DownSocket");

                        // Если не удалось прочитать все данные,
                        // скорее всего, клиент отвалился,
                        // защитим сервер и не будем передавать 
                        // ему половинчатые данные.
                        return;
                    }
                    LogNetwork(8, "End ReceiveExact 2nd time from DownSocket");

                    transaction.Stage++; // 4

                    // Закрываем сокет на чтение, 
                    // т. к. больше читать из него не собираемся
                    LogNetwork(9, "Shutdown DownSocket (Receive)");
                    downSocket.Shutdown(SocketShutdown.Receive);

                    // Полностью сформированный клиентский запрос
                    byte[] request = requestStorage.ToArray();
                    transaction.Request = request;

                    if (!CheckTheRequest(transaction))
                    {
                        // Пакет не прошел проверку
                        return;
                    }

                    // Преобразуем запрос, если необходимо
                    TransformTheRequest
                        (
                            transaction,
                            ref request
                        );

                    transaction.Stage++; // 5

                    // Создаем сокет для связи с сервером
                    Socket upSocket = new Socket
                        (
                            AddressFamily.InterNetwork,
                            SocketType.Stream,
                            ProtocolType.Tcp
                        );
                    LogNetwork
                        (
                            10,
                            "Got UpSocket"
                        );

                    try
                    {
                        IPEndPoint endPoint = new IPEndPoint
                            (
                                RemoteIP,
                                RemotePort
                            );

                        try
                        {
                            LogNetwork
                                (
                                    11, 
                                    "Begin Connect UpSocket: {0}",
                                    endPoint
                                );
                            upSocket.Connect(endPoint);
                        }
                        catch (Exception ex)
                        {
                            LogNetwork(12, "Fail Connect UpSocket");
                            WriteException(ex);

                            // Если не удалось связаться с сервером,
                            // все последующие действия не имеют смысла
                            WakeTheSystemAdministrator(transaction);
                            return;
                        }
                        LogNetwork(13, "End Connect UpSocket");

                        transaction.Stage++; // 6

                        // Настраиваем сокет для наилучшей связи
                        LogNetwork(14, "Setup UpSocket");
                        SetupSocket(upSocket);

                        // Замеряем время, затраченное на обработку
                        // собственно сервером ИРБИС.
                        Stopwatch irbisStopwatch = new Stopwatch();
                        irbisStopwatch.Start();

                        // Посылаем запрос на сервер
                        LogNetwork(15, "Begin SendAll UpSocket");
                        if (!SendAll(upSocket, request))
                        {
                            LogNetwork(16, "Fail SendAll UpSocket");
                            // Если не удалось отправить данные на сервер,
                            // бросаем это гиблое дело
                            WakeTheSystemAdministrator(transaction);
                            return;
                        }
                        LogNetwork(17, "End SendAll UpSocket");

                        transaction.Stage++; // 7

                        //// Закрываем сокет на посылку,
                        //// т. к. больше посылать на сервер ничего не будем
                        //upSocket.Shutdown(SocketShutdown.Send);

                        MemoryStream responseStorage = new MemoryStream();
                        LogNetwork(18, "Begin ReceiveAll UpSocket");
                        if (!ReceiveAll
                            (
                                upSocket,
                                responseStorage
                            ))
                        {
                            LogNetwork(19, "Fail ReceiveAll UpSocket");
                            WakeTheSystemAdministrator(transaction);
                            return;
                        }
                        LogNetwork(20, "End ReceiveAll UpSocket");

                        transaction.Stage++; // 8

                        irbisStopwatch.Stop();
                        transaction.IrbisDuration = irbisStopwatch.ElapsedMilliseconds;

                        // Закрываем сокет на прием
                        //LogNetwork(21, "Shutdown UpSocket (Receive)");
                        //upSocket.Shutdown(SocketShutdown.Receive);

                        // Полностью сформированный ответ сервера
                        byte[] response = responseStorage.ToArray();
                        transaction.Response = response;

                        if (!CheckTheResponse(transaction))
                        {
                            // Пакет не прошел проверку
                            return;
                        }

                        transaction.Stage++; // 9

                        LogNetwork(22, "Begin SendAll DownSocket");
                        if (!SendAll
                            (
                                downSocket,
                                response
                            ))
                        {
                            LogNetwork(23, "Fail SendAll DownSocket");
                            return;
                        }
                        LogNetwork(24, "End SendAll DownSocket");

                        transaction.Stage++; // 10

                        //LogNetwork(25, "Shutdown DownSocket (Send)");
                        //downSocket.Shutdown(SocketShutdown.Send);

                        transaction.Stage++; // 11
                    }
                    finally
                    {
                        LogNetwork(26, "Close UpSocket");
                        upSocket.Close(10);
                        transaction.Stage++; // 12
                    }

                }
                finally
                {
                    LogNetwork(27, "Close DownSocket");
                    downSocket.Close(10);
                    transaction.Stage++; // 13
                }
            }
            catch (Exception ex)
            {
                WriteException(ex);
                Debug.WriteLine(ex);
            }
            finally
            {
                roundtripStopwatch.Stop ();
                transaction.RoundtripDuration 
                    = roundtripStopwatch.ElapsedMilliseconds;

                LogTransaction(transaction);

                Interlocked.Decrement(ref Load);
            }
        }

        // ====================================================================

        /// <summary>
        /// Прослушивание соединения.
        /// </summary>
        public static void ListenerRoutine()
        {
            try
            {
                ListenerSocket = new Socket
                    (
                        AddressFamily.InterNetwork,
                        SocketType.Stream,
                        ProtocolType.Tcp
                    );

                IPEndPoint endPoint = new IPEndPoint
                    (
                        IPAddress.Any,
                        LocalPort
                    );

                ListenerSocket.Bind
                    (
                        endPoint
                    );

                WriteDump
                    (
                        "Listening on {0}",
                        ListenerSocket.LocalEndPoint
                    );

                ListenerSocket.Listen
                    (
                        Backlog
                    );

                while (!Stop)
                {
                    Socket downSocket = ListenerSocket.Accept();

                    // Создаем поток на каждое клиентское соединение
                    Thread handlerThread = new Thread
                        (
                        HandlerRoutine
                        )
                        {
                            Name = "HandlerThread",
                            IsBackground = true
                        };

                    handlerThread.Start
                        (
                            downSocket
                        );
                }
            }
            catch (Exception ex)
            {
                WriteException(ex);
                Debug.WriteLine(ex);
            }

            // Ожидаем завершения операций логирования
            while (Engine.IsBusy)
            {
                Thread.Sleep(100);
            }
        }

        // ====================================================================

        public static void StartListener()
        {
            Thread thread = new Thread
                (
                    ListenerRoutine
                )
            {
                Name = "ListenerThread",
                IsBackground = true
            };

            thread.Start();
        }
    }

    // ====================================================================
}
